/*
 * Copyright (c) 2018 Razeware LLC
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * Notwithstanding the foregoing, you may not use, copy, modify, merge, publish,
 * distribute, sublicense, create a derivative work, and/or sell copies of the
 * Software in any work that is designed, intended, or marketed for pedagogical or
 * instructional purposes related to programming, coding, application development,
 * or information technology.  Permission for such use, copying, modification,
 * merger, publication, distribution, sublicensing, creation of derivative works,
 * or sale is expressly withheld.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

package com.raywenderlich.podplay.repository

import android.arch.lifecycle.LiveData
import com.raywenderlich.podplay.db.PodcastDao
import com.raywenderlich.podplay.model.Episode
import com.raywenderlich.podplay.model.Podcast
import com.raywenderlich.podplay.service.FeedService
import com.raywenderlich.podplay.service.RssFeedResponse
import com.raywenderlich.podplay.util.DateUtils
import kotlinx.coroutines.experimental.CommonPool
import kotlinx.coroutines.experimental.android.UI
import kotlinx.coroutines.experimental.launch

class PodcastRepo(private var feedService: FeedService,
                  private var podcastDao: PodcastDao) {

  fun getPodcast(feedUrl: String, callback: (Podcast?) -> Unit) {

    launch(CommonPool) {

      val podcast = podcastDao.loadPodcast(feedUrl)

      if (podcast != null) {
        podcast.id?.let {
          podcast.episodes = podcastDao.loadEpisodes(it)
          launch(UI){
            callback(podcast)
          }
        }
      } else {

        feedService.getFeed(feedUrl, { feedResponse ->

          if (feedResponse == null) {
            launch(UI) {
              callback(null)
            }
          } else {
            val podcast = rssResponseToPodcast(feedUrl, "", feedResponse)
            launch(UI) {
              callback(podcast)
            }
          }
        })
      }
    }

  }

  fun getAll(): LiveData<List<Podcast>>
  {
    return podcastDao.loadPodcasts()
  }

  fun updatePodcastEpisodes(callback: (List<PodcastUpdateInfo>) -> Unit) {

    val updatedPodcasts: MutableList<PodcastUpdateInfo> = mutableListOf()

    val podcasts = podcastDao.loadPodcastsStatic()

    var processCount = podcasts.count()
    for (podcast in podcasts) {
      getNewEpisodes(podcast, { newEpisodes ->

        if (newEpisodes.count() > 0) {
          saveNewEpisodes(podcast.id!!, newEpisodes)
          updatedPodcasts.add(PodcastUpdateInfo(podcast.feedUrl, podcast.feedTitle, newEpisodes.count
          ()))
        }
        processCount--

        if (processCount == 0) {
          callback(updatedPodcasts)
        }
      })
    }
  }

  private fun getNewEpisodes(localPodcast: Podcast, callBack: (List<Episode>) -> Unit) {

    feedService.getFeed(localPodcast.feedUrl, { response ->

      if (response != null) {
        val remotePodcast = rssResponseToPodcast(localPodcast.feedUrl, localPodcast.imageUrl, response)
        remotePodcast?.let {

          val localEpisodes = podcastDao.loadEpisodes(localPodcast.id!!)

          val newEpisodes = remotePodcast.episodes.filter { episode ->
            localEpisodes.find { episode.guid == it.guid } == null
          }

          callBack(newEpisodes)
        }
      } else {
        callBack(listOf())
      }
    })
  }

  private fun saveNewEpisodes(podcastId: Long, episodes: List<Episode>) {
    launch(CommonPool) {
      for (episode in episodes) {
        episode.podcastId = podcastId
        podcastDao.insertEpisode(episode)
      }
    }
  }

  fun save(podcast: Podcast) {
    launch(CommonPool) {
      val podcastId = podcastDao.insertPodcast(podcast)
      for (episode in podcast.episodes) {
        episode.podcastId = podcastId
        podcastDao.insertEpisode(episode)
      }
    }
  }

  fun delete(podcast: Podcast) {
    launch(CommonPool) {
      podcastDao.deletePodcast(podcast)
    }
  }

  private fun rssResponseToPodcast(feedUrl: String, imageUrl: String, rssResponse:
  RssFeedResponse): Podcast? {

    val items = rssResponse.episodes ?: return null
    val description = if (rssResponse.description == "") rssResponse.summary else rssResponse.description

    return Podcast(null, feedUrl, rssResponse.title, description, imageUrl,
        rssResponse.lastUpdated, episodes = rssItemsToEpisodes(items))
  }

  private fun rssItemsToEpisodes(episodeResponses: List<RssFeedResponse.EpisodeResponse>): List<Episode> {
    return episodeResponses.map {
      Episode(
          it.guid ?: "",
          null,
          it.title ?: "",
          it.description ?: "",
          it.url ?: "",
          it.type ?: "",
          DateUtils.xmlDateToDate(it.pubDate),
          it.duration ?: ""
      )
    }
  }

  class PodcastUpdateInfo (val feedUrl: String, val name: String, val newCount: Int)
}